module convection

  use vtype_module

  implicit none

  public :: ConvCalcRadTendency
  public :: ConvCalcRadConvTendency

contains

  !----------------------------------------------------------------------------

  subroutine ConvCalcCalcDTDH( &
    & kmax, r_Press, r_Temp,   & ! (in)
    & r_DTempDEnthalpy         & ! (out)
    & )

#ifdef THERMOPROP_VIRACpLSF
    use thermo_prop_co2_viracplsf, only : &
#else
    use thermo_prop_co2_sw1996, only : &
#endif
      & ThermoPropDens,           &
      & ThermoPropCp

    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press         (0:kmax)
    real(dp), intent(in ) :: r_Temp          (0:kmax)
    real(dp), intent(out) :: r_DTempDEnthalpy(0:kmax)


    ! Local variables
    !
    real(dp) :: Dens
    integer  :: k


    !$omp parallel do private(k,Dens) shared(kmax,r_Press,r_Temp,r_DTempDEnthalpy)
    do k = 0, kmax
      Dens          = ThermoPropDens( r_Press(k), r_Temp(k) )

      r_DTempDEnthalpy(k) = 1.0d0 / ThermoPropCp( Dens, r_Temp(k) )
    end do
    !$omp end parallel do


  end subroutine ConvCalcCalcDTDH

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadTendency(                                 &
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp,                                      & ! (in)
    & r_Enthalpy,                                                 & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelEnthalpy, DelSurfTemp                                  & ! (out)
    & )

    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press        (0:kmax)
    real(dp), intent(in ) :: r_Temp         (0:kmax)
    real(dp), intent(in ) :: r_Enthalpy     (0:kmax)
    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux    (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux    (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux(0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux(0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelEnthalpy  (0:kmax)
    real(dp), intent(out) :: DelSurfTemp


    ! Local variables
    !
    real(dp) :: z_DelAtmMass (1:kmax)
    real(dp) :: r_RadFlux    (0:kmax)
    real(DP) :: ra_DelRadFlux(0:kmax,-1:NDelFlxMax)

    real(dp) :: r_DTempDEnthalpy(0:kmax)

    real(dp) :: rr_Mat(-1:kmax,-1:kmax)
    real(dp) :: r_Vec (-1:kmax)
    integer  :: k
    integer  :: kr
    integer  :: kc


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do

    r_RadFlux     = r_RadUwFlux     - r_RadDwFlux
    ra_DelRadFlux = ra_DelRadUwFlux - ra_DelRadDwFlux


!!$    call ConvCalcConvT2H( &
!!$      & kmax, r_Press, r_Temp,  &
!!$      & r_Enthalpy              &
!!$      & )
    r_DTempDEnthalpy = -1.0d100
    call ConvCalcCalcDTDH( &
      & kmax, r_Press, r_Temp,   & ! (in)
      & r_DTempDEnthalpy         & ! (out)
      & )


    rr_Mat = 0.0d0
    r_Vec  = 0.0d0
    !
    ! Ts
    kr = -1
    !
    kc = -1
    rr_Mat(kr,kc) = SurfHeatCap / DelTime + ra_DelRadFlux(0,kc)
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = + ra_DelRadFlux(0,kc) * r_DTempDEnthalpy(kc)
    end do
    !
    r_Vec(kr)     = - r_RadFlux(0)
    !
    ! T (k=0)
    !   T0 = ( T2 - T1 ) / ( 2 - 1 ) * ( 0 - 1 ) + T1
    !      = - ( T2 - T1 ) + T1
    !      = 2 * T1 - T2
    !   T0 - 2 * T1 + T2 = 0
    !   r_Temp(0) + r_DelTemp(0) - 2 * ( r_Temp(1) + r_DelTemp(1) )
    !     + r_Temp(2) + r_DelTemp(2) = 0
    !   r_DelTemp(0) - 2 * r_DelTemp(1) + r_DelTemp(2) 
    !     = - ( r_Temp(0) - 2 * r_Temp(1) + r_Temp(2) )
    !   r_DTDH(0) * r_DelH(0) - 2 * r_DTDH(1) * r_DelH(1) + r_DTDH(2) * r_DelH(2) 
    !     = - ( r_Temp(0) - 2 * r_Temp(1) + r_Temp(2) )
    kr = 0
    kc =   0
    rr_Mat(kr,kc) =           r_DTempDEnthalpy(kc)
    kc =   1
    rr_Mat(kr,kc) = - 2.0d0 * r_DTempDEnthalpy(kc)
    kc =   2
    rr_Mat(kr,kc) =           r_DTempDEnthalpy(kc)
    !
    r_Vec(kr)     = - ( r_Temp(0) - 2.0d0 * r_Temp(1) + r_Temp(2) )
    !
    ! T
    do kr = 1, kmax
      kc = -1
      rr_Mat(kr,kc) = ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc)
      do kc = 0, NDelFlxMax
        rr_Mat(kr,kc) = ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) ) &
          &             * r_DTempDEnthalpy(kc)
      end do
      kc = kr
      rr_Mat(kr,kc) = rr_Mat(kr,kc) + z_DelAtmMass(kr) / DelTime
      !
      r_Vec(kr) = - ( r_RadFlux(kr) - r_RadFlux(kr-1) )
    end do

    call ludecomp_prep_simple ( kmax+2, rr_Mat )
    call ludecomp_solve_simple( kmax+2, rr_Mat, r_Vec )

    ! r_Vec(-1) = surface temperature
    ! r_Vec(0 ) = enthalpy just above the surface
    ! r_Vec(1 ) = enthalpy at 1st atmospheric level
    ! r_Vec(2 ) = enthalpy at 2nd atmospheric level
    ! ...

    DelSurfTemp  = r_Vec(-1)

    r_DelEnthalpy = r_Vec(0:kmax)

!!$    r_Enthalpy   = r_Enthalpy + r_Vec(0:kmax)
!!$    call ConvCalcConvH2T(    &
!!$      & kmax, r_Press, r_Enthalpy, &
!!$      & r_TempA                    &
!!$      & )
!!$    r_DelTemp   = r_TempA - r_Temp


  end subroutine ConvCalcRadTendency

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadConvTendencyCoreAll(                      &
    & z_FlagConvection,                                           & ! (in)
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp, SurfTemp,                            & ! (in)
    & r_Enthalpy, r_DTempDEnthalpy,                               & ! (in)
    & z_VertDelEnthalpyUppLevel,                                  & ! (in)
    & z_VertDelEnthalpyLowLevel,                                  & ! (in)
    & z_DVertDelEnthalpyDTempUppLevel,                            & ! (in)
    & z_DVertDelEnthalpyDTempLowLevel,                            & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelEnthalpy, DelSurfTemp, SurfSensFlux,                   & ! (out)
    & NConvLayer,                                                 & ! (out)
    & a_ksConvLayer, a_keConvLayer                                & ! (out)
    & )

    ! Full level index (mid-poing of layers)
    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    logical , intent(in ) :: z_FlagConvection(1:kmax)
    real(dp), intent(in ) :: r_Press          (0:kmax)
    real(dp), intent(in ) :: r_Temp           (0:kmax)
    real(dp), intent(in ) :: SurfTemp
    real(dp), intent(in ) :: r_Enthalpy       (0:kmax)
    real(dp), intent(in ) :: r_DTempDEnthalpy (0:kmax)

    real(DP), intent(in ) :: z_VertDelEnthalpyUppLevel      (1:kmax)
    real(DP), intent(in ) :: z_VertDelEnthalpyLowLevel      (1:kmax)
    real(DP), intent(in ) :: z_DVertDelEnthalpyDTempUppLevel(1:kmax)
    real(DP), intent(in ) :: z_DVertDelEnthalpyDTempLowLevel(1:kmax)

    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux      (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux      (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux  (0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux  (0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelEnthalpy    (0:kmax)
    real(dp), intent(out) :: DelSurfTemp
    real(DP), intent(out) :: SurfSensFlux
    integer , intent(out) :: NConvLayer
    integer , intent(out) :: a_ksConvLayer(0:kmax)
    integer , intent(out) :: a_keConvLayer(0:kmax)


    ! Function
    !
!!$    real(DP) :: LOCALPressH2OVapSat
!!$    real(DP) :: LOCALDPressH2OVapSatDTemp

    ! Local variables
    !
    real(dp) :: z_DelAtmMass (1:kmax)
    real(dp) :: r_RadFlux    (0:kmax)
    real(DP) :: ra_DelRadFlux(0:kmax,-1:NDelFlxMax)
    real(dp) :: rr_Mat(-2:kmax,-2:kmax)
    real(dp) :: r_Vec (-2:kmax)

    real(dp) :: TotalDEnthalpyDt
    real(dp) :: FluxTOA
    real(dp) :: FluxSfc
    real(dp) :: Residual

    integer  :: kConvS
    integer  :: kConvE

    integer  :: k
    integer  :: kr
    integer  :: kc


    r_RadFlux     = r_RadUwFlux     - r_RadDwFlux
    ra_DelRadFlux = ra_DelRadUwFlux - ra_DelRadDwFlux


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do


    ! check layer thickness of convection from the surface
    kConvS = 1
    do k = kConvS+1, kmax
      if ( .not. z_FlagConvection(k) ) then
        exit
      end if
    end do
    kConvE = k - 1

    a_ksConvLayer(0) = kConvS
    a_keConvLayer(0) = kConvE


    rr_Mat = 0.0d0
    r_Vec  = 0.0d0
    !
    ! Energy conservation
    kr = -2
    kc = -2 ! sensible heat flux
    rr_Mat(kr,kc) = - 1.0d0
    kc = -1
    rr_Mat(kr,kc) = + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) )
    !
!!$    kc =  0
!!$    rr_Mat(kr,kc) = + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) ) &
!!$      & * r_DTempDEnthalpy(kc)
!!$    kc =  1
!!$    rr_Mat(kr,kc) = + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) ) &
!!$      & * r_DTempDEnthalpy(kc)
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = &
        & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) ) &
        &   * r_DTempDEnthalpy(kc)
    end do
    !
    do kc = 1, kConvE
      ! memo : Latent heat is neglected to conserve energy
      rr_Mat(kr,kc) = rr_Mat(kr,kc) &
        & + z_DelAtmMass(kc) / DelTime
    end do
    !
    r_Vec(kr) = - ( r_RadFlux(kConvE) - r_RadFlux(0) )

    !
    ! Ts
    kr = -1
    kc = -2
    rr_Mat(kr,kc) = + 1.0d0
    kc = -1
    rr_Mat(kr,kc) = + SurfHeatCap / DelTime + ra_DelRadFlux(0,kc)
    !
!!$    kc =  0
!!$    rr_Mat(kr,kc) = + ra_DelRadFlux(0,kc) &
!!$      & * r_DTempDEnthalpy(kc)
!!$    kc =  1
!!$    rr_Mat(kr,kc) = + ra_DelRadFlux(0,kc) &
!!$      & * r_DTempDEnthalpy(kc)
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = rr_Mat(kr,kc) &
        & + ra_DelRadFlux(0,kc)     &
        &   * r_DTempDEnthalpy(kc)
    end do
    !
    r_Vec(kr) = - r_RadFlux(0)
    !
    ! T (k=0)
    !   T0 = Ts
    !   r_Temp(0) + r_DelTemp(0) = SurfTemp + DelSurfTemp
    !   - DelSurfTemp + r_DelTemp(0) = SurfTemp - r_Temp(0)
    !   - DelSurfTemp + r_DTempDEnthalpy * r_DelEnthalpy(0)
    !     = SurfTemp - r_Temp(0)
    kr = 0
    kc =  -1
    rr_Mat(kr,kc) = - 1.0d0
    kc =   0
    rr_Mat(kr,kc) = + 1.0d0 * r_DTempDEnthalpy(kc)
    !
    r_Vec(kr)     = SurfTemp - r_Temp(0)
    !
    ! T (1 <= k < kConvE)
    do kr = 1, kConvE
      ! VDeltaEnthalpy : Vertical enthalpy difference between two layers
      ! VDeltaEnthalpy
      !   =   z_VertDelEnthalpyLowLevel(k)
      !         + z_DVertDelEnthalpyDTempLowLevel(k) * r_DelTemp(k-1)
      !     + z_VertDelEnthalpyUppLevel(k)
      !         + z_DVertDelEnthalpyDTempUppLevel(k) * r_DelTemp(k  )
      !   =   z_VertDelEnthalpyLowLevel(k)
      !         + z_DVertDelEnthalpyDTempLowLevel(k)
      !             * r_DTempDEnthalpy(k-1) * r_DelEnthalpy(k-1)
      !     + z_VertDelEnthalpyUppLevel(k)
      !         + z_DVertDelEnthalpyDTempUppLevel(k)
      !             * r_DTempDEnthalpy(k  ) * r_DelEnthalpy(k  )
      !--------
      ! r_Enthalpy(k-1) + r_DelEnthalpy(k-1) + VDeltaEnthalpy
      ! = r_Enthalpy(k) + r_DelEnthalpy(k)
      !--------
      ! - r_DelEnthalpy(k-1) + r_DelEnthalpy(k)
      ! = r_Enthalpy(k-1) - r_Enthalpy(k) + VDeltaEnthalpy
      !--------
      ! - ( 1 + z_DVertDelEnthalpyDTempLowLevel(k) * r_DTempDEnthalpy(k-1) )
      !     * r_DelEnthalpy(k-1)
      ! + ( 1 - z_DVertDelEnthalpyDTempUppLevel(k) * r_DTempDEnthalpy(k  ) )
      !     * r_DelEnthalpy(k  )
      ! = r_Enthalpy(k-1) - r_Enthalpy(k)
      !     + z_VertDelEnthalpyLowLevel(k) + z_VertDelEnthalpyUppLevel(k)
      !
      kc = kr-1
      rr_Mat(kr,kc) = - ( 1.0d0 + z_DVertDelEnthalpyDTempLowLevel(kr) * r_DTempDEnthalpy(kr-1) )
      kc = kr
      rr_Mat(kr,kc) =   ( 1.0d0 - z_DVertDelEnthalpyDTempUppLevel(kr) * r_DTempDEnthalpy(kr  ) )
      !
      r_Vec(kr) = &
        &   r_Enthalpy(kr-1) - r_Enthalpy(kr) &
        & + z_VertDelEnthalpyLowLevel(kr) + z_VertDelEnthalpyUppLevel(kr)
    end do


    ! convection in atmosphere

    NConvLayer = 0

    k = kConvE + 1
    loop_conv_in_atm : do
      if ( z_FlagConvection(k) ) then
        kConvS = k
        find_top_of_conv : do kConvE = kConvS+1, kmax
          if ( .not. z_FlagConvection(kConvE) ) then
            exit find_top_of_conv
          end if
        end do find_top_of_conv
        kConvE = kConvE-1

        NConvLayer = NConvLayer + 1
        a_ksConvLayer(NConvLayer) = kConvS
        a_keConvLayer(NConvLayer) = kConvE

        ! energy conservation
        kr = kConvS
        kc = -1
        rr_Mat(kr,kc) = &
          & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(kConvS-1,kc) )
        !
        do kc = 0, NDelFlxMax
          rr_Mat(kr,kc) = &
            & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(kConvS-1,kc) ) &
            &   * r_DTempDEnthalpy(kc)
        end do
        !
        do kc = kConvS, kConvE
          rr_Mat(kr,kc) = rr_Mat(kr,kc) &
            & + z_DelAtmMass(kc) / DelTime
        end do
        !
        r_Vec(kr) = - ( r_RadFlux(kConvE) - r_RadFlux(kConvS-1) )

        do kr = kConvS+1, kConvE
          ! same as convection layer which originates from the surface
          kc = kr-1
          rr_Mat(kr,kc) = &
            & - ( 1.0d0 + z_DVertDelEnthalpyDTempLowLevel(kr) * r_DTempDEnthalpy(kr-1) )
          kc = kr
          rr_Mat(kr,kc) = &
            &   ( 1.0d0 - z_DVertDelEnthalpyDTempUppLevel(kr) * r_DTempDEnthalpy(kr  ) )
          !
          r_Vec(kr) = &
            &   r_Enthalpy(kr-1) - r_Enthalpy(kr) &
            & + z_VertDelEnthalpyLowLevel(kr) + z_VertDelEnthalpyUppLevel(kr)
        end do

        k = kConvE + 1

      else
        ! stable layer
        kr = k

        kc = -1
        rr_Mat(kr,kc) = + ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) )
        !
        do kc = 0, NDelFlxMax
          rr_Mat(kr,kc) = rr_Mat(kr,kc) &
            & + ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) ) &
            &   * r_DTempDEnthalpy(kc)
        end do
        !
        kc = kr
        rr_Mat(kr,kc) = rr_Mat(kr,kc) &
          & + z_DelAtmMass(kc) / DelTime
        !
        r_Vec(kr) = - ( r_RadFlux(kr) - r_RadFlux(kr-1) )

        k = k + 1

      end if
      if ( k > kmax ) then
        exit loop_conv_in_atm
      end if
    end do loop_conv_in_atm

    call ludecomp_prep_simple ( (kmax+3), rr_Mat )
    call ludecomp_solve_simple( (kmax+3), rr_Mat, r_Vec )
    !
    r_DelEnthalpy = r_Vec(0:kmax)
    !
    DelSurfTemp   = r_Vec(-1)
    !
    SurfSensFlux  = r_Vec(-2)


    ! check energy conservation
    !
    TotalDEnthalpyDt = 0.0d0
    do k = 1, kmax
      TotalDEnthalpyDt = TotalDEnthalpyDt &
        & + r_DelEnthalpy(k) * z_DelAtmMass(k)
    end do
    TotalDEnthalpyDt = TotalDEnthalpyDt / DelTime
    !
    FluxTOA = r_RadFlux(kmax)
    k = -1
    FluxTOA = FluxTOA + ra_DelRadFlux(kmax,k) * DelSurfTemp
    do k = 0, NDelFlxMax
      FluxTOA = FluxTOA &
        & + ra_DelRadFlux(kmax,k) * r_DTempDEnthalpy(k) * r_DelEnthalpy(k)
    end do
    !
    FluxSfc = r_RadFlux(0)
    k = -1
    FluxSfc = FluxSfc + ra_DelRadFlux(0,k) * DelSurfTemp
    do k = 0, NDelFlxMax
      FluxSfc = FluxSfc &
        & + ra_DelRadFlux(0,k) * r_DTempDEnthalpy(k) * r_DelEnthalpy(k)
    end do
    !
    ! TotalDelEnthalpy = - ( FluxTOA - FluxSfc ) + SurfSensFlux
    Residual = TotalDEnthalpyDt + ( FluxTOA - FluxSfc ) - SurfSensFlux
    !
    if ( abs( Residual / FluxSfc ) > 1.0d-10 ) then
      write( 6, * ) 'Residual may be large : ', &
        & TotalDEnthalpyDt, FluxTOA, FluxSfc, SurfSensFlux, Residual
      call flush( 6 )
    end if


  end subroutine ConvCalcRadConvTendencyCoreAll

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadConvTendency(                             &
    & FlagMoistConvection,                                        & ! (in)
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp, SurfTemp, r_VMRH2OVap, r_MeanMolWt,  & ! (in)
    & r_Enthalpy,                                                 & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelEnthalpy, DelSurfTemp, r_DelVMRH2OVap, SurfSensFlux,   & ! (out)
    & FlagStdOut                                                  & ! (in)
    & )

    logical , intent(in ) :: FlagMoistConvection
    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press        (0:kmax)
    real(dp), intent(in ) :: r_Temp         (0:kmax)
    real(dp), intent(in ) :: SurfTemp
    real(dp), intent(in ) :: r_VMRH2OVap    (0:kmax)
    real(DP), intent(in ) :: r_MeanMolWt    (0:kmax)
    real(DP), intent(in ) :: r_Enthalpy     (0:kmax)
    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux    (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux    (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux(0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux(0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelEnthalpy  (0:kmax)
    real(dp), intent(out) :: DelSurfTemp
    real(DP), intent(out) :: r_DelVMRH2OVap (0:kmax)
    real(DP), intent(out) :: SurfSensFlux
    logical , intent(in ), optional :: FlagStdOut


    ! Local variables
    !
    real(dp) :: r_EnthalpyTentative            (0:kmax)
    real(dp) :: r_DTempDEnthalpy               (0:kmax)

    real(DP) :: r_DEnthalpyDlnPAd              (0:kmax)
    real(DP) :: r_DDEnthalpyDlnPDTempAd        (0:kmax)
    real(DP) :: z_VertDelEnthalpyUppLevel      (1:kmax)
    real(DP) :: z_VertDelEnthalpyLowLevel      (1:kmax)
    real(DP) :: z_DVertDelEnthalpyDTempUppLevel(1:kmax)
    real(DP) :: z_DVertDelEnthalpyDTempLowLevel(1:kmax)

    real(DP) :: z_RadFluxConv  (1:kmax)

    ! Full level index (mid-poing of layers)
    integer , parameter :: kConvSfcS = 1
    integer             :: kConvSfcE

    real(dp)            :: z_DelAtmMass (1:kmax)

    real(dp)            :: Press

    integer             :: iItr
    integer             :: NItr = 1000
    logical             :: z_FlagConvection    (1:kmax)
    logical             :: z_FlagConvectionPrev(1:kmax)
    logical             :: FlagFinish

    integer             :: NConvLayer
    integer             :: a_ksConvLayer(0:kmax)
    integer             :: a_keConvLayer(0:kmax)

    integer             :: k


    if ( FlagMoistConvection ) then
      write( 6, * ) 'ERROR : FlagMoistConvection of true is not supported, now.'
      stop
    end if


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do


    ! Find top of convection
    do k = 1, kmax
      z_RadFluxConv(k) =                                  &
        & - (     r_RadUwFlux(k  ) - r_RadDwFlux(k  )     &
        &     - ( r_RadUwFlux(k-1) - r_RadDwFlux(k-1) ) )
    end do
    do kConvSfcE = 1, kmax
      if ( z_RadFluxConv(kConvSfcE) < 0.0d0 ) exit
    end do
    if ( kConvSfcE > kmax ) then
      write( 6, * ) 'Convection seems to reach top of model from the surface'
      stop
    end if


    r_DTempDEnthalpy = -1.0d100
    call ConvCalcCalcDTDH( &
      & kmax, r_Press, r_Temp,   & ! (in)
      & r_DTempDEnthalpy         & ! (out)
      & )
    call ConvCalcDryConvDEnthalpyDlnP(  &
      & kmax, r_Press, r_Temp,          & ! (in )
      & r_DEnthalpyDlnPAd,              & ! (out)
      & r_DDEnthalpyDlnPDTempAd         & ! (out)
      & )

    do k = 1, kmax
      if ( r_Press(k) == 0.0d0 ) then
        Press = sqrt( r_Press(k-1) * r_Press(k-1) * 0.9d0 )
      else
        Press = sqrt( r_Press(k-1) * r_Press(k)           )
      end if
      !
      ! These values mean value at upper level minus lower level.
      ! Negative sign mean upper level value (enthalpy) is smaller than
      ! lower level value.
      !
      z_VertDelEnthalpyUppLevel(k) = &
        & - r_DEnthalpyDlnPAd(k  ) * log( Press        / r_Press(k) )
      z_VertDelEnthalpyLowLevel(k) = &
        & - r_DEnthalpyDlnPAd(k-1) * log( r_Press(k-1) / Press      )
      !
      z_DVertDelEnthalpyDTempUppLevel(k) = &
        & - r_DDEnthalpyDlnPDTempAd(k  ) * log( Press        / r_Press(k) )
      z_DVertDelEnthalpyDTempLowLevel(k) = &
        & - r_DDEnthalpyDlnPDTempAd(k-1) * log( r_Press(k-1) / Press      )
    end do


    ! Initialization
    z_FlagConvection = .false.

    iteration : do iItr = 1, NItr

      r_EnthalpyTentative = r_Enthalpy + r_DelEnthalpy


      z_FlagConvectionPrev = z_FlagConvection

      ! check unstable layers
      z_FlagConvection(kConvSfcS:kConvSfcE) = .true.
      do k = kConvSfcE+1, kmax
!!$        if (   r_EnthalpyTentative(k-1)+z_VertDelEnthalpy(k  ) &
!!$          &  > r_EnthalpyTentative(k  ) ) then
        if (   r_EnthalpyTentative(k-1)         &
          &      + z_VertDelEnthalpyLowLevel(k) &
          &      + z_VertDelEnthalpyUppLevel(k) &
          &  > r_EnthalpyTentative(k  ) ) then
          ! Unstable
          z_FlagConvection(k-1) = .true.
          z_FlagConvection(k  ) = .true.
        end if
      end do
      FlagFinish = .true.
      loop_check_finish : do k = 1, kmax
        if ( z_FlagConvectionPrev(k) .neqv. z_FlagConvection(k) ) then
          FlagFinish = .false.
          exit loop_check_finish
        end if
      end do loop_check_finish
      if ( FlagFinish ) then
        exit iteration
      end if


      ! calc tendency
      call ConvCalcRadConvTendencyCoreAll(                            &
        & z_FlagConvection,                                           &
        & Grav, SurfHeatCap, DelTime,                                 &
        & kmax, r_Press, r_Temp, SurfTemp,                            &
        & r_Enthalpy, r_DTempDEnthalpy,                               &
        & z_VertDelEnthalpyUppLevel,                                  & ! (in)
        & z_VertDelEnthalpyLowLevel,                                  & ! (in)
        & z_DVertDelEnthalpyDTempUppLevel,                            & ! (in)
        & z_DVertDelEnthalpyDTempLowLevel,                            & ! (in)
        & NDelFlxMax,                                                 &
        & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, &
        & r_DelEnthalpy, DelSurfTemp, SurfSensFlux,                   &
        & NConvLayer,                                                 & ! (out)
        & a_ksConvLayer, a_keConvLayer                                & ! (out)
        & )

    end do iteration
    if ( iItr >= NItr ) then
      write( 6, * ) 'ERROR : convection does not finish.'
      stop
    end if
    if ( present( FlagStdOut ) ) then
      if ( FlagStdOut ) then
        write( 6, * ) 'Convection iterations    : ', iItr
        write( 6, * ) 'Convection from surface  : ', a_keConvLayer(0), 'layers'
        write( 6, * ) 'Convection in atmosphere : ', NConvLayer, 'layers'
        do k = 1, NConvLayer
          write( 6, '(a,i3.3,a,i3.3)' ) '   ', &
            & a_ksConvLayer(k), '-', a_keConvLayer(k)
        end do
        call flush( 6 )
      end if
    end if


    r_DelVMRH2OVap = 0.0d0


  end subroutine ConvCalcRadConvTendency

  !----------------------------------------------------------------------------

  subroutine ConvCalcDryConvDEnthalpyDlnP( &
    & kmax, r_Press, r_Temp,               & ! (in )
    & r_DEnthalpyDlnP,                     & ! (out)
    & r_DDEnthalpyDlnPDTemp                & ! (out)
    & )

#ifdef THERMOPROP_VIRACpLSF
    use thermo_prop_co2_viracplsf, only : &
#else
    use thermo_prop_co2_sw1996, only : &
#endif
      & ThermoPropDens,                 &
      & ThermoPropDPressDDensConstTemp, &
      & ThermoPropDPressDTempConstDens

    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: r_Press              (0:kmax)
    real(DP), intent(in ) :: r_Temp               (0:kmax)
    real(DP), intent(out) :: r_DEnthalpyDlnP      (0:kmax)
    real(DP), intent(out) :: r_DDEnthalpyDlnPDTemp(0:kmax)


    ! Local variables
    !
    real(dp) :: Press
    real(dp) :: Temp
    real(dp) :: Dens
    real(dp) :: DPDTCRho
    real(dp) :: DPDRhoCT
    integer  :: k


    do k = 0, kmax
      Press = r_Press(k)
      Temp  = r_Temp (k)
      Dens  = ThermoPropDens( Press, Temp )
      r_DEnthalpyDlnP(k) = Press / Dens

      DPDTCRho = ThermoPropDPressDTempConstDens( Dens, Temp )
      DPDRhoCT = ThermoPropDPressDDensConstTemp( Dens, Temp )
      r_DDEnthalpyDlnPDTemp(k) = &
        & - r_DEnthalpyDlnP(k) / Dens * ( - DPDTCRho / DPDRhoCT )
    end do


  end subroutine ConvCalcDryConvDEnthalpyDlnP

  !----------------------------------------------------------------------------

end module convection
