!------------------------------------------------------------------------------

subroutine CalcHeight( &
  & kmax,                                   &
  & r_Press, r_Temp, r_VMRH2O, r_MeanMolWt, &
  & r_Height                                &
  & )

  use vtype_module

  use constants0, only : &
    & GasRUniv

  use rad_planetconst, only : &
    & Grav,            &
    & MolWtH2O

  implicit none

  integer , intent(in ) :: kmax
  real(dp), intent(in ) :: r_Press    (0:kmax)
  real(dp), intent(in ) :: r_Temp     (0:kmax)
  real(dp), intent(in ) :: r_VMRH2O   (0:kmax)
  real(dp), intent(in ) :: r_MeanMolWt(0:kmax)
  real(dp), intent(out) :: r_Height   (0:kmax)


  ! Local variables
  !
  real(dp) :: r_GasR   (0:kmax)
  real(dp) :: r_EpsV   (0:kmax)
  real(dp) :: r_MMRH2O (0:kmax)
  real(dp) :: r_VirTemp(0:kmax)
  real(dp) :: PressTop
  integer  :: k


  r_GasR = GasRUniv / r_MeanMolWt
  r_EpsV = MolWtH2O / r_MeanMolWt

  r_MMRH2O = r_VMRH2O * MolWtH2O / r_MeanMolWt

  r_VirTemp = r_Temp &
    & * ( 1.0d0 + ( ( ( 1.0d0 / r_EpsV ) - 1.0d0 ) * r_MMRH2O ) )

  ! dp/dz = -rho * g = - p * g / ( R * T )
  ! 1/p * dp/dz = - g / ( R * T )
  ! d(lnp)/dz = - g / ( R * T )
  ! dz = - R * T / g * d(lnp)

  r_Height(0) = 0.0d0
  do k = 0+1, kmax
    if ( r_Press(k) <= 0.0d0 ) then
      PressTop = r_Press(k-1) * 1.0d-1
    else
      PressTop = r_Press(k)
    end if
    r_Height(k) = r_Height(k-1)                              &
      & -   ( r_GasR(k-1) + r_GasR(k) ) / 2.0d0              &
      &   * ( r_VirTemp(k-1) + r_VirTemp(k) ) / 2.0d0        &
      &   / Grav                                             &
      &   * log( PressTop / r_Press(k-1) )
  end do


end subroutine CalcHeight

!------------------------------------------------------------------------------

subroutine FilterUsingEnthalpy( &
  & FilterCoef,    & ! (in)
  & kmax, r_Press, & ! (in)
  & r_Enthalpy     & ! (inout)
  & )

  use vtype_module

  use rad_planetconst, only : &
    & Grav

  implicit none

  real(DP), intent(in   ) :: FilterCoef
  integer , intent(in   ) :: kmax
  real(DP), intent(in   ) :: r_Press      (0:kmax)
  real(DP), intent(inout) :: r_Enthalpy   (0:kmax)


  ! Local variables
  !
  real(DP) :: z_DelAtmMass  (1:kmax)
  real(DP) :: r_EnthalpyFlux(0:kmax)
  real(DP) :: r_DelEnthalpy (0:kmax)

  integer  :: k


  do k = 1, kmax
    z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
  end do

  k = 0
  r_EnthalpyFlux(k) = 0.0d0
  do k = 0+1, kmax-1
    r_EnthalpyFlux(k) = &
      & - FilterCoef * ( z_DelAtmMass(k+1) + z_DelAtmMass(k) ) / 2.0d0 &
      &   * ( r_Enthalpy(k+1) - r_Enthalpy(k) )
  end do
  k = kmax
  r_EnthalpyFlux(k) = 0.0d0
  !
  k = 0
  r_DelEnthalpy(k) = 0.0d0
  do k = 0+1, kmax
    r_DelEnthalpy(k) = &
      & - ( r_EnthalpyFlux(k) - r_EnthalpyFlux(k-1) ) / z_DelAtmMass(k)
  end do
  !
  r_Enthalpy = r_Enthalpy + r_DelEnthalpy


end subroutine FilterUsingEnthalpy

!------------------------------------------------------------------------------

subroutine ConvCalcConvT2H( &
  & kmax, r_Press, r_Temp,  &
  & r_Enthalpy              &
  & )

  use vtype_module

#ifdef THERMOPROP_VIRACpLSF
  use thermo_prop_co2_viracplsf, only : &
#else
  use thermo_prop_co2_sw1996, only : &
#endif
    & ThermoPropDens,       &
    & ThermoPropEnthalpy

  implicit none

  integer , intent(in ) :: kmax
  real(dp), intent(in ) :: r_Press        (0:kmax)
  real(dp), intent(in ) :: r_Temp         (0:kmax)
  real(dp), intent(out) :: r_Enthalpy     (0:kmax)


  ! Local variables
  !
  real(dp) :: Dens
  integer  :: k


  !$omp parallel do private(k,Dens) shared(kmax,r_Press,r_Temp,r_Enthalpy)
  do k = 0, kmax
    Dens          = ThermoPropDens( r_Press(k), r_Temp(k) )
    r_Enthalpy(k) = ThermoPropEnthalpy( Dens, r_Temp(k) )
  end do
  !$omp end parallel do


end subroutine ConvCalcConvT2H

!----------------------------------------------------------------------------

subroutine ConvCalcConvH2T(    &
  & kmax, r_Press, r_Enthalpy, &
  & r_Temp                     &
  & )

  use vtype_module

#ifdef THERMOPROP_VIRACpLSF
  use thermo_prop_co2_viracplsf, only : &
#else
  use thermo_prop_co2_sw1996, only : &
#endif
    & ThermoPropDensTempFromPressEnthalpy

  implicit none

  integer , intent(in ) :: kmax
  real(dp), intent(in ) :: r_Press        (0:kmax)
  real(dp), intent(in ) :: r_Enthalpy     (0:kmax)
  real(dp), intent(out) :: r_Temp         (0:kmax)


  ! Local variables
  !
  real(dp) :: Dens
  integer  :: k


  !$omp parallel do private(k,Dens) shared(kmax,r_Press,r_Temp,r_Enthalpy)
  do k = 0, kmax
    call ThermoPropDensTempFromPressEnthalpy( &
      & r_Press(k), r_Enthalpy(k), & ! (in)
      & Dens, r_Temp(k)            & ! (out)
      & )
  end do
  !$omp end parallel do


end subroutine ConvCalcConvH2T

!------------------------------------------------------------------------------

